/* =====================================================================================
 * ACTO - Motor Control for TUIO
 * =====================================================================================
 * controls two independent electro motors connected to H-Bridge version 1.0 - 3.1
 * with "electronic stability control" feature
 *
 * date: 21. Feb 2014
 * version: v2.2
 *
 * project: ACTO
 * authors: Emanuel Vonach, Georg Gerstweiler
 * Interactive Media Systems Group, Vienna University of Technology, Austria
 * www.ims.tuwien.ac.at
 * Contact: vonach@ims.tuwien.ac.at, gerstweiler@ims.tuwien.ac.at
 * =====================================================================================
 */

/* input int steeringAngle, int overallSpeed
  steeringAngle 0-255: imagine as circle or steering wheel
  0       = ++ straight ahead
  1-31    = ++ slight right turn, right motor less than left one
  32      = +0 right turn with right motor stop
  33-63   = +- right turn with right motor help in counter direction
  64      = +- right turn on spot, rightmotor = -leftmotor
  65-95   = -+ right turn with right motor help in counter direction, but in reverse
  96      = -0 right turn with right motor stop, but in reverse
  97-127  = -- slight right turn, right motor less than left one, but in reverse
  128     = -- straight reverse (with speed 0 active brake)
  129-159 = -- slight left turn, left motor less than right one, but in reverse
  160     = 0- left turn with left motor stop, but in reverse
  161-191 = +- left turn with left motor help in counter direction, but in reverse
  192     = -+ left turn on spot, leftmotor = -rightmotor
  193-223 = -+ left turn with left motor help in counter direction
  224     = 0+ left turn with left motor stop
  225-255 = ++ slight left turn, left motor less than right one
*/


/*
direct motorcontrol:
    D   S
    0   0: engine idle (much less power consumption!!)
    0   200-255: forward
    255 0-55: backward
    255 255: active brake
*/

//handicaps of motor units:
#ifdef MOTOR_DEFAULT                      //default values
const float maxFwd[] = {255,255};         //define handicap of motors and max Speed
const float minSpeed = 90;                //define speed threshhold: 0+minSpeed = stop = 255-minSpeed
const int directionPreset = 1;            //define where "front" is for motor unit
#endif
#ifdef MOTOR_ALICE                        //TUIO 1
const float maxFwd[] = {180,255};         //define handicap of motors and max Speed  {R, L}
const float minSpeed = 100;                //define speed threshhold: 0+minSpeed = stop = 255-minSpeed
const int directionPreset = 1;            //define where "front" is for motor unit
#endif
#ifdef MOTOR_BOB                          //TUIO 2
const float maxFwd[] = {225,255};         //define handicap of motors and max Speed
const float minSpeed = 55;                //define speed threshhold: 0+minSpeed = stop = 255-minSpeed
const int directionPreset = 1;            //define where "front" is for motor unit
#endif
#ifdef MOTOR_CHARLIE                      //TUIO 3
const float maxFwd[] = {255,220};         //define handicap of motors and max Speed  {L, R}
const float minSpeed = 100;                //define speed threshhold: 0+minSpeed = stop = 255-minSpeed
const int directionPreset = -1;            //define where "front" is for motor unit
#endif
#ifdef MOTOR_DAVE                         //TUIO 4
const float maxFwd[] = {255,255};         //define handicap of motors and max Speed
const float minSpeed = 55;                //define speed threshhold: 0+minSpeed = stop = 255-minSpeed
const int directionPreset = -1;            //define where "front" is for motor unit
#endif
#ifdef MOTOR_ERIN                         //TUIO 5
const float maxFwd[] = {255,255};         //define handicap of motors and max Speed
const float minSpeed = 55;                //define speed threshhold: 0+minSpeed = stop = 255-minSpeed
const int directionPreset = -1;            //define where "front" is for motor unit
#endif
#ifdef MOTOR_FRANK                        //TUIO 6
const float maxFwd[] = {255,255};         //define handicap of motors and max Speed
const float minSpeed = 55;                //define speed threshhold: 0+minSpeed = stop = 255-minSpeed
const int directionPreset = 1;            //define where "front" is for motor unit
#endif


const int motor_L1 = 3;    // =D3 PWM  0: PNP11 leitet, NPN11 sperrt.  255:PNP11 sperrt, NPN11 leitet.  PWM!!!!
const int motor_L2 = 18;   // =A4=18   0: vorwärts: PNP12 leitet, NPN12 sperrt.  255: rückwärts: PNP12 sperrt, NPN12 leitet.  kein PWM!!!

const int motor_R1 = 10;   // =D10 PWM 0: PNP21 leitet, NPN21 sperrt. 255:PNP21 sperrt, NPN21 leitet.  PWM!!!!
const int motor_R2 = 19;   // =A5=19   0: vorwärts: PNP22 leitet, NPN22 sperrt.  255: rückwärts: PNP22 sperrt, NPN22 leitet.  kein PWM!!!

// helper to hold curent state for ESC
float currLeftSpeed = 0;     // PWM duty cycle, sign is direction
float currRightSpeed = 0;    // PWM duty cycle, sign is direction

// for driveMS
unsigned long driveStampEnd = 0;

// for ESC
const unsigned long escSpanSteps = 5;          // timespan of step in ms
const unsigned long escSpan = escSpanSteps*4;  // complete time of ESC, should be multiple of escSpanSteps & even

const int escThr = 100;          // threshold for speed jump when ESC should be engaged
int escStepCnt = 0;
unsigned long escStamp = 0;

float leftEscSteps = 0;
float rightEscSteps = 0;

void initMotorUnit(){
  //set motor idle
  driveDirectionSpeed(0,0);
}

void motorLoop() {
  //check if Motor should stop because timestamp has passed
  if (driveStampEnd != 0 && millis() >= driveStampEnd) {
    driveStampEnd = 0;
    driveDirectionSpeed(0,0);
  }
  //check if currently in process of ESC
  if (escStepCnt > 0) doEscStep(); // if ESC is active 
}

//drive until stopped, public access
void driveDirectionSpeed(byte steerAngle, byte overallSpeed) {
  driveStampEnd = 0;
  driveDirectionSpeed((int)steerAngle, (int)overallSpeed);
}

//drive for specific ms, public access
void driveDirectionSpeed(byte steerAngle, byte overallSpeed, int ms) {
  driveStampEnd = millis() + (long)ms;
  driveDirectionSpeed((int)steerAngle, (int)overallSpeed);
}

//private method for driving
void driveDirectionSpeed(int steerAngle, int overallSpeed) {
  
  //check for invalid data
  if (steerAngle >= 256 || overallSpeed >= 256) {
    errorHandler(MOTORERROR);
    return;
  }
  
  //special case: avoid active brake for too long, so end after 100ms
  if (steerAngle == 128 && overallSpeed == 0) driveStampEnd = millis() + 100;
  
  float leftFactor = 1.0;
  float rightFactor = 1.0;
  
  if (steerAngle <= 64) {              // left motor = +speed, right motor = +speed(0) >= 0 >= -speed(64)
    rightFactor = (32 - (float)steerAngle)/32;
  }
  else if (steerAngle <= 128) {        // left motor = -speed, right motor = +speed(65) > 0 >= -speed(128)
    leftFactor = -1.0;
    rightFactor = (96 - (float)steerAngle)/32;
  }
  else if (steerAngle < 192) {        // left motor = -speed(129) > 0 <= +speed(192), right motor = -speed  //missing "=" on purpose
    leftFactor = -1*(160 - (float)steerAngle)/32;
    rightFactor = -1.0;
  }
  else if (steerAngle <= 255) {        // left motor = -speed(193) > 0 <= +speed(255), right motor = speed
    leftFactor = -1*(224 - (float)steerAngle)/32;
  }

  float newLeftSpeed = 0;
  float newRightSpeed = 0;
  
  // adapt speed of each motor as fraction of overall speed, respective to steerAngle (directionPreset +-1 to invert direction)
  newLeftSpeed = (float)overallSpeed * leftFactor * directionPreset;    
  newRightSpeed = (float)overallSpeed * rightFactor * directionPreset;
  
// JUST DEBUG:
//  Serial.print(" steerAngle:   ");
//  Serial.println(steerAngle);
//  Serial.print(" speed: ");
//  Serial.println(overallSpeed);
//  Serial.println(" ");
//  
//  Serial.println("current: ");
//  Serial.print(currLeftSpeed);
//  Serial.print(" - ");
//  Serial.print(currRightSpeed);
//  Serial.println(" ");
//  
//  Serial.println("new: ");
//  Serial.print(newLeftSpeed);
//  Serial.print(" - ");
//  Serial.print(newRightSpeed);
//  Serial.println(" ");
  
  //check if speed change too great
  int diffL = currLeftSpeed - newLeftSpeed;
  if (diffL < 0) diffL = diffL*-1;
  
  int diffR = currRightSpeed - newRightSpeed;
  if (diffR < 0) diffR = diffR*-1;
  
  // if change to great, drive with ESC to avoid voltage spikes or slippery
  if (diffL >= escThr || diffR >= escThr) driveESC(newLeftSpeed, newRightSpeed);
  else drive(newLeftSpeed, newRightSpeed);
}

// electronic stability control ESC, to avoid hard direction changes (+255 to -255) and slipping
//void driveESC(int newLeftDirection, float newLeftSpeed, int newRightDirection, float newRightSpeed) {
void driveESC(float newLeftSpeed, float newRightSpeed) {
  
  //calculate number of steps
  escStepCnt = escSpan/escSpanSteps;
  
  //calculate difference
  float leftEscDiff = currLeftSpeed - newLeftSpeed;
  float rightEscDiff = currRightSpeed - newRightSpeed;
  
  // calculate step size
  leftEscSteps = leftEscDiff/escStepCnt;
  rightEscSteps = rightEscDiff/escStepCnt;

  escStamp = millis()-escSpanSteps;

// JUST DEBUG:  
//  Serial.println("step cnt: ");
//  Serial.println(escStepCnt);
//  Serial.println("steps");
//  Serial.println(leftEscSteps);
//  Serial.println(rightEscSteps);
//  Serial.println(" ");
}

void doEscStep() {
  if (millis() - escStamp >= escSpanSteps) {   //every escSpanSteps ms
    //reduce / increase speed by EscSteps
    currLeftSpeed = currLeftSpeed - leftEscSteps;
    currRightSpeed = currRightSpeed - rightEscSteps;

    drive(currLeftSpeed, currRightSpeed);

    escStepCnt = escStepCnt - 1;
    escStamp = millis();
  }
}


// direct motorcontrol for both motorunits
// Parameters: direction and speed for each motor // direction: 0=reverse, 1=forward, speed: 0=stop to 255=maxSpeed
//void drive(int newLeftDirection, float newLeftSpeed, int newRightDirection, float newRightSpeed){
void drive(float newLeftSpeed, float newRightSpeed){

  int leftSpeed = 0;
  int rightSpeed = 0;
  int newLeftDirection = 0;
  int newRightDirection = 0;
  
//  String debugOut;
  
  //save globally for ESC
  currLeftSpeed = newLeftSpeed;
  currRightSpeed = newRightSpeed;
  
  //calculate handicap
  newLeftSpeed = maxFwd[0] * newLeftSpeed/255;
  newRightSpeed = maxFwd[1] * newRightSpeed/255;

  // inverts speed depending on direction
  if(newLeftSpeed < 0) {
    //newLeftDirection = 0;
    
    // newLeftSpeed is negative -> thus the -
    leftSpeed = (int)ceil(255 - minSpeed - minSpeed * newLeftSpeed / 255) ;  //0-255 == (maxFwd[0]-minSpeed) to maxFwd[0] (e.g. 200-255)  //ceil -> speed 1 drives min speed
    
//    debugOut = String("R_") + String((int)ceil(newLeftSpeed)) + String("_") + String(leftSpeed);
  } else {
    newLeftDirection = 1;
    
    leftSpeed = (int)(minSpeed - minSpeed * newLeftSpeed / 255);  //0-255 == minSpeed to 0 (e.g. (55-0)  //floor -> speed 1 drives min speed
    
//    debugOut = String("V_") + String((int)floor(newLeftSpeed)) + String("_") + String(leftSpeed);
  }
  
  if(newRightSpeed < 0) {
    //newRightDirection = 0;
    
    // newRightSpeed is negative -> thus the -
    rightSpeed = (int)ceil(255 - minSpeed - minSpeed * newRightSpeed / 255) ;  //0-255 == (maxFwd[0]-minSpeed) to maxFwd[0] (e.g. 200-255)  //ceil -> speed 1 drives min speed
    
//    debugOut += String("R_") + String((int)ceil(newRightSpeed)) + String("_") + String(rightSpeed);
  } else {
    newRightDirection = 1;
    
    rightSpeed = (int)(minSpeed - minSpeed * newRightSpeed / 255);  //0-255 == minSpeed to 0 (e.g. (55-0)  //floor -> speed 1 drives min speed
    
//    debugOut += String("V_") + String((int)floor(newRightSpeed)) + String("_") + String(rightSpeed);
  }

// JUST DEBUG:
//  Serial.println(debugOut);  //remove debug String construction when ready!!

//  Serial.println("step: ");
//  Serial.print(newLeftSpeed);
//  Serial.print(" - ");
//  Serial.print(newRightSpeed);
//  Serial.println(" ");

  // if motor should stop, then rather stop it completely to prevent damage
  if (leftSpeed >= minSpeed && leftSpeed <= (255-minSpeed)) {
    if (newLeftDirection > 0) leftSpeed = 255;
    else leftSpeed = 0;
  }
  if (rightSpeed >= minSpeed && rightSpeed <= (255-minSpeed)) {
    if (newRightDirection > 0) rightSpeed = 255;
    else rightSpeed = 0;
  }

  // give actual motor command
  analogWrite(motor_L1, leftSpeed);      //duty cycle 0-55 bzw. 200-255
  analogWrite(motor_L2, newLeftDirection*255);   //0: vorwärts: PNP12 leitet, NPN12 sperrt.  255: rückwärts: PNP12 sperrt, NPN12 leitet.
  
  analogWrite(motor_R1, rightSpeed);      //duty cycle 0-55 bzw. 200-255
  analogWrite(motor_R2, newRightDirection*255);   //0: vorwärts: PNP22 leitet, NPN22 sperrt.  255: rückwärts: PNP22 sperrt, NPN22 leitet.
}






